package PetshopSpider;

import java.io.*;
import java.net.*;
import java.util.*;


/*
  Class HTMLRetriever is concurrent part of SmallSpider.  
  */

public class HTMLRetriever extends Thread {

  // Constructor
  HTMLRetriever() {

    // Not ready to accet jobs, yet.
    id = next_id++;
    this.setDaemon(true);
  }

  public static void init(SpiderSetting sset, JobManager jm, URLChecker uc) {
    next_id = 0;
    job_manager = jm;
    url_checker = uc;
    timeout = sset.timeout;
    max_level = sset.max_level;
    robots_txt = sset.robots_txt;
    domains = sset.domains;
    extensions = sset.extensions;
    ignore_servers = sset.ignore_servers;
    mimes = sset.mimes;
    default_files = sset.default_files;
    disable_extensions = sset.disable_extensions;
  }

  // Run
  public void run() {

    String response;
    InputStream in;
    BufferedInputStream bin; BufferedOutputStream bout;

    for(;;) {
      // Ready to accept job.
      Print.debug("HR: Notify finish to JobManager  ID:"+id);

      // Clean up
      target_url = null; www_host = null;
      url_str = null; host_str = null;
      extension_str = null; filename_str = null;
      response = null; in = null;
      bin = null; bout = null;
      cm = null; key = null;

      job_manager.finish(id);
      suspend(); // Waiting next job.

      // Process a job.
      count = target_url.getCount();

      // A level exceeds maximum level.
      if(count>max_level) {
	notifyFinish(RETRIEVE_SKIP);
	continue;
      }

      // Preparation for accessing URLChecker.
      url_str = target_url.getURLString();
      host_str = target_url.getHost();
      extension_str = target_url.getExtension();
      filename_str = target_url.getFileName();
      key = target_url.getKey();

      Print.debug("HR: Check URL ID:"+id);
      int host_state = url_checker.getHostState(host_str);
      if(host_state==WWWHost.HOST_IS_DOWN) {
	notifyFinish(RETRIEVE_SKIP);
	continue;
      }
      if(host_state==WWWHost.HOST_IS_BUSY) {
	notifyFinishAndBack(RETRIEVE_SKIP);
	continue;
      }
      if(url_checker.inRobotsTxt(host_str, url_str)) {
	notifyFinish(RETRIEVE_SKIP);
	continue;
      }
      if(url_checker.isAlreadyRetrieved(key)) {
	notifyFinish(RETRIEVE_SKIP);
	continue;
      }

      Print.verbose("HR: Create Conenction ID:"+id);

      // Invoke connection maker. If cm exceeds timeout secs, a thread is joined.
      cm = new ConnectionMaker(target_url);
      cm.start();
      try {
	cm.join(60000);
      } catch(InterruptedException e) {}
      cm.stop();

      // Get InputStream from ConnectionMaker.
      Print.verbose("HR: Get "+target_url.getURLString()+" ID:"+id);

      // If returned value is null, it is failed.
      if(!cm.isSuccess()) {
	Print.debug("HR: ConnectionMaker failed ID:"+id);
	// Notify connection is failed.
	notifyFinishAndBack(RETRIEVE_FAIL);
	continue;
      }

      response = cm.getResponse();
      in = cm.getInputStream();

      // Maybe, this connection is failed.
      if(response==null || in==null) {
	Print.debug("HR: ConnectionMaker returns nulls ID:"+id);
	notifyFinish(RETRIEVE_FAIL);
	continue;
      }

      Print.verbose("HR: Server response "+response+" ID:"+id);
      if(response.indexOf("200")<0) {
	// Sendback if this URL has more possible URLs.

	if(target_url.getHasNext()) {
	  target_url.nextURL();
	  notifyFinishAndBack(RETRIEVE_SKIP);
	  continue;
	}

	notifyFinish(RETRIEVE_SKIP);
	continue;
      }


      // Check writable if target_url is writable URL or not.
      boolean writable = hasEnableExtension(extension_str);
      // MIME extension check.
      if(!writable && !target_url.isEndsWithDirectory() && isEnableMIME(cm.getContentType()))
	  writable = true;

      // Prepare invokcation URLExtractor
      bin = new BufferedInputStream(in);

      try {

	if(writable) {
	  File f = new File(target_url.getFile().getParent());
	  if(!f.isDirectory())
	    f.mkdirs();
	  bout = new BufferedOutputStream
	    (new FileOutputStream(target_url.getFile()));
	  ue = new URLExtractor(bin, bout, target_url);     
	} else
	  ue = new URLExtractor(bin, target_url);
      
	// Get URLs. Invoke URLExtractor.
	Print.debug("HR: Call URLExtractor ID:"+id);
	ue.start();
	try {
	  ue.join(300000);
	} catch(InterruptedException e) {}
	ue.stop();
      
	if(!ue.isSuccess()) {
	  Print.debug("HR: URLextractor abnormally finished. ID:"+id);
	  job_manager.addJob(target_url);
	} else {
	  addJobs(ue.getURLs());
	  Print.debug("HR: URLExtractor is finished. ID:"+id);
	}

	try {
	  bin.close();
	} catch(IOException e) {}
	try {
	  if(writable && bout!=null)
	    bout.close();
	} catch(IOException e) {}

	notifyFinish(RETRIEVE_SUCCESS);

      } catch (IOException g) {
	System.err.println("HR: Write file error occured at "+url_str+","+target_url.getFile().getPath()+" ID:"+id);

	// Temporary.
	notifyFinish(RETRIEVE_SUCCESS);
      }
    }
  }

  // Interfaces methods. called by WWWHost.
  public synchronized void setJob(WWWHost wh, Resource r) {
    Print.debug("HR: setJob is invoked for "+r.getURLString());
    target_url = r;
    www_host = wh;
    this.resume();
  }

  // Adding a job to JobManager's buffer.
  private void addJobs(Enumeration e) {

    String host_str, url_str, file_str, extension;
    int cnt;
    Resource rs;
    Integer key;

    while(e.hasMoreElements()) {
      rs = (Resource)e.nextElement();
      host_str = rs.getHost();
      url_str = rs.getURLString();
      file_str = rs.getFileName();
      extension = rs.getExtension();
      key = rs.getKey();

      cnt = count+1;

      // Check reject-conditions
      if(hasDisableExtension(extension)) continue;
      if(!inDomain(host_str)) continue;
      if(isIgnore(host_str)) continue;
      if(!checkProtocol(rs)) continue;

      // Use URL checker
      int host_state = url_checker.getHostState(host_str);
      if(host_state==WWWHost.HOST_IS_DOWN) continue;
      if(robots_txt)
	if(url_checker.inRobotsTxt(host_str, url_str)) continue;
      if(url_checker.isAlreadyRetrieved(key)) continue;

      // A case like http://www.aaa.bbb/a/
      if(rs.isEndsWithDirectory() || rs.isWithoutExtension()) {
	Print.debug("HR: Set has_next property for "+url_str+" ID:"+id);
	rs.setHasNext();
	rs.nextURL();
	addJob(rs);
	continue;
      }
    
      // If writable extension. The conditions progress and save are same now.
      if(!hasDisableExtension(extension)) 
	addJob(rs);
    }
  }


  // Final method to throw jobs.
  private void addJob(Resource rs) {
    Print.debug("HR: Throw URL to buffer  URL:"+rs.getURLString()+" ID:"+id);
    job_manager.addJob(rs);
  }

  // Notify URL process is finished.
  private void notifyFinish(int state) {
    Print.debug("HR: HR notifies finish to WWWHost state:"+state+" ID:"+id);
    url_checker.notifyFinish(target_url.getKey());
    www_host.notifyFinish(state);
  }

  // Doesn't notify to URLchecker. Send back a url.
  private void notifyFinishAndBack(int state) {
    Print.debug("HR: HR notifies finish and back to WWWHost state:"+state+" ID:"+id);
    job_manager.addJob(target_url);
    www_host.notifyFinish(state);
  }

  // Check domain
  private boolean inDomain(String host) {
    for(int i=0;i<domains.length;i++)
      if(host.endsWith(domains[i])) return true;
    return false;
  }

  private boolean isIgnore(String host) {
    for(int i=0;i<ignore_servers.length;i++)
      if(host.compareTo(ignore_servers[i])==0)
	return true;
    return false;
  }

  private boolean hasEnableExtension(String extension) {
    Print.debug("HR:Extension test ID:"+id);
    // extetnsion includes lower case string only.
    String ext = extension.toLowerCase();
    for(int i=0;i<extensions.length;i++)
      if(ext.compareTo(extensions[i])==0)
	return true;
    return false;
  }

  private boolean hasDisableExtension(String extension) {
    // disabe_extensions includes lower case string only.
    String ext = extension.toLowerCase();
    for(int i=0;i<disable_extensions.length;i++)
      if(ext.compareTo(disable_extensions[i])==0)
	return true;
    return false;
  }

  // Check MIME type, it goes to save, return true.
  private boolean isEnableMIME(String content_type) {
    if(content_type==null) return false;

    for(int i=0;i<mimes.length;i++)
      if(content_type.indexOf(mimes[i])>=0)
	return true;
    return false;
  }
  
  private boolean checkProtocol(Resource rs) {
    Print.debug("HR: Protocol test ID:"+id);
    String s = rs.getProtocol();
    if(s.compareTo("http")==0) return true;
    return false;
  }

  // Variables
  int id; // Identifier of this object.
  Resource target_url; // Information about current target URL.

  ConnectionMaker cm; // Inner class for making connection with exactly timeout.
  URLExtractor ue; // URLExtractor object;
  WWWHost www_host; // For callbacks.

  // Statics
  private static JobManager job_manager; // Job Manager object.
  private static URLChecker url_checker; // URLChecker object.
  private static int timeout, max_level;
  private static boolean robots_txt;
  private static String[] domains, extensions, ignore_servers, mimes,
    default_files, disable_extensions;
  public static int next_id;

  // Constants
  public final static int RETRIEVE_SKIP = 0;
  public final static int RETRIEVE_FAIL = 1;
  public final static int RETRIEVE_SUCCESS = 2;

  // Temporary
  int count;
  String url_str, host_str, extension_str, filename_str;
  Integer key;
}








