package PetshopSpider;

import java.net.*;
import java.util.*;

class JobManager extends Thread {
  // Constructor
  JobManager(SpiderSetting sset) {
    // Initialization
    this.sset = sset;

    // Initialize HTMLRetrievers activity array.
    waiting = new boolean[sset.number_of_retrievers];
    for(int i=0;i<waiting.length;i++) {
      waiting[i] = false;
    }

    // Initialize HTMLRetriever array.
    retrievers = new HTMLRetriever[sset.number_of_retrievers];
    number_of_waiting = 0;
    number_of_requests = 0;
    number_of_busy = 0;
    number_of_stable = 0;

    this.timeout = sset.timeout;
    this.urls = sset.urls;

    // Initialize connection related arrays.
    number_of_hosts = 0;
    hosts = new WWWHost[NUMBER_OF_SERVERS];
    requestHR = new boolean[NUMBER_OF_SERVERS];
    host_table = new Hashtable();

    start_search_point = 0;

    // For debug.
    active_status = new StringBuffer();
    host_status = new StringBuffer(60);
  }

  // For mutually reference structure.
  public void setURLChecker(URLChecker uc) {
    url_checker = uc;
  }

  public void run() {
    // JobManager run first.
    setPriority(Thread.MIN_PRIORITY);

    // For stating URLs.
    Print.debug("JM: Setup initial resources.");
    WWWHost wh = new WWWHost("PETSHOP SPECIAL");
    addHost("PETSHOP SPECIAL",wh);

    // Start first WWWHost. And waiting to start.
    wh.start();
    try {
      Thread.sleep(1000);
    } catch(InterruptedException e){}

    for(int i=0;i<urls.length;i++)
      wh.addJob(urls[i]);

    // Create HTMLRetrievers
    for(int i=0;i<retrievers.length;i++)
      retrievers[i] = new HTMLRetriever();

    for(int i=0;i<retrievers.length;i++) {
      retrievers[i].setPriority(Thread.MAX_PRIORITY);
      retrievers[i].start();
    }

    wh = null;

    // Wait until all Retrievers are active.
    Print.verbose("JM:Wait all retrievers are active");
    while(number_of_waiting!=retrievers.length)
      suspend();

    int request_id, hr_id;

    Print.verbose("JM:start main loop");

    requested = true;

    for(;;) {

      Print.verbose("JM: Number of waitings "+number_of_waiting);
      printStatus();

      // Check finish all job.

      Print.verbose("JM: Number of queues requests HR "+number_of_requests);
      Print.verbose("JM: Number of busy hosts "+number_of_busy);

      // Wait next job.
      if(!requested)
	suspend();
      requested = false;

      if(number_of_stable<number_of_hosts)
	continue;

      if(number_of_waiting==retrievers.length  && number_of_requests==0 && number_of_busy==0)
	break;

      if(number_of_waiting==0)
	continue;

      if(number_of_requests==0)
	continue;

      hr_id = findWaiting();
      request_id = findRequest();

      Print.debug("JM: Throw job using HR ID:"+hr_id);
      
      throw_job(hr_id, request_id);
    }
  }

  // Check number_of_busy hosts busy hosts.
  public synchronized void notifyBusy() {
    Print.debug("JM: notifyBusy");
    number_of_busy++;
  }

  public synchronized void recoverFromBusy() {
    Print.debug("JM: recoverFromBusy");
    number_of_busy--;
    requested = true;
    this.resume();
  }

  // Callback from WWWHost. Request/Cancel HR.
  public synchronized void requestHR(int i) {
    requestHR[i] = true;
    number_of_requests++;
    requested = true;
    this.resume();
  }

  public synchronized void cancelHR(int i) {
    requestHR[i] = false;
    number_of_requests--;
    requested = true;
    this.resume();
  }

  // Callback from WWWHost. Report stable/instable.
  public synchronized void notifyStable() {
    number_of_stable++;
    requested = true;
    resume();
  }

  public synchronized void notifyInStable() {
    number_of_stable--;
  }

  // Callback from HTMLRetrievers.
  public synchronized void finish(int i) {
    waiting[i] = true;
    number_of_waiting++;
    requested = true;
    this.resume();
  }
 
  public synchronized void addJob(Resource rs) {
    String host_str = rs.getHost();
    String url_str = rs.getURLString();
    WWWHost wh = (WWWHost)host_table.get(host_str);
    wh.addJob(rs);
  }

  // Add WWWHost objects.
  public synchronized void addHost(String url_str, WWWHost wh) {
    // Expand array if no space is avilable.
    if(number_of_hosts == hosts.length) {
      WWWHost[] temp_www = new WWWHost[number_of_hosts*2];
      for(int i=0;i<number_of_hosts;i++)
	temp_www[i] = hosts[i];
      hosts = temp_www; temp_www = null;
      boolean[] temp_bool = new boolean[number_of_hosts*2];
      for(int i=0;i<number_of_hosts;i++)
	temp_bool[i] = requestHR[i];
      requestHR = temp_bool; temp_bool = null;
    }
    requestHR[number_of_hosts] = false;
    hosts[number_of_hosts++] = wh;
    host_table.put(url_str, wh);
    requested = true;
    this.resume();
  }

  // Inner methods

  // Send HTMLRetriever to WWWHost. True for need jobs more.
  private synchronized void throw_job(int hr_id, int request_id) {
    waiting[hr_id] = false;
    requestHR[request_id] = false;
    number_of_waiting--;
    number_of_requests--;
    hosts[request_id].giveHR(retrievers[hr_id]);
  }

  // Find out an usable HTMLRetriever.
  private int findWaiting() {
    for(int i=0;i<retrievers.length; i++)
      if(waiting[i]) return i;
    return -1;
  }

  // Find out request from WWWHost.
  private int findRequest() {
    start_search_point++;
    for(int i=0;i<number_of_hosts; i++) {
      if(start_search_point>=number_of_hosts)
	start_search_point = 0;
      if(requestHR[start_search_point])
	return start_search_point;
      start_search_point++;
    }
    return -1;
  }

  // For debug usage.
  private void printStatus() {
    active_status.setLength(0);
    host_status.setLength(0);
    for(int i=0; i<waiting.length; i++)
      if(waiting[i])
	active_status.append('!');
      else
	active_status.append('-');

    for(int i=0; i<number_of_hosts; i++)
      if(requestHR[i])
	host_status.append('!');
      else
	host_status.append('-');
    Print.verbose("Thread state:"+active_status+" Hosts state:"+host_status+" Stable state:"+number_of_stable);

  }

  /*
    Variables
    */

  private boolean[] waiting; // State of HTMLRetrievers
  private HTMLRetriever[] retrievers;
  private int number_of_waiting, number_of_requests, number_of_stable;

  private WWWHost[] hosts; // The array of hosts.
  private boolean[] requestHR; // Flag array for request HR from WWWHost.
  private Hashtable host_table;
  private int number_of_hosts;
  private int number_of_busy; // Number of busy hosts.
  private int start_search_point; // Start point for findRequest.

  private SpiderSetting sset;
  private URLChecker url_checker;
  private int timeout;
  private Resource[] urls;

  private StringBuffer active_status, host_status;
  
  private final static int BUFFER_DEFAULT_SIZE = 500;
  private final static int NUMBER_OF_SERVERS = 60;

  private boolean requested;
}

