package PetshopSpider;

import java.net.*;
import java.io.*;
import java.util.*;

class Resource {
  // Static
  private static String directory;
  private static String[] default_files;
  private static Random rnd;
  private static long rnd_seed;

  // Variables
  URL url;
  String path;
  File file;
  String url_str;
  boolean ok;
  int counter;

  boolean has_next;
  String origin_url;
  int next_state;

  Resource(String url_str) {
    try {
      url = new URL(new String(url_str));
      String s = url.getFile();
      file = new File("/"+url.getHost()+url.getFile());
      has_next = false;
      ok = true;
    } catch(MalformedURLException e) {
      System.err.println("RS:MalformedURLException");
      ok = false;
    }
  }

  Resource(String url_str, int cnt) {
    try {
      url = new URL(new String(url_str));
      String s = url.getFile();
      file = new File(directory+"/"+url.getHost()+url.getFile());
      counter = cnt;
      has_next = false;
      ok = true;
    } catch(MalformedURLException e) {
      System.err.println("RS:MalformedURLException");
      ok = false;
    }
  }

  Resource(Resource rs, String url_str, int cnt) {
    try {
      url = new URL(rs.getURL(),new String(url_str));
      String s = url.getFile();
      file = new File(directory+"/"+url.getHost()+url.getFile());
      counter = cnt;
      has_next = false;
      ok = true;
    } catch(MalformedURLException e) {
      System.err.println("RS:MalformedURLException");
      ok = false;
    }
  }

  Resource(Resource rs, String url_str) {
    try {
      url = new URL(rs.getURL(),new String(url_str));
      String s = url.getFile();
      file = new File(directory+"/"+url.getHost()+url.getFile());
      has_next = false;
      ok = true;
    } catch(MalformedURLException e) {
      Print.debug("RS: MalformedURLException "+url_str);
      ok = false;
    }
  }

  // Set static variable directory. 
  public static void init(SpiderSetting sset) {
    directory = sset.storage_directory;
    Resource.default_files = sset.default_files;
    // Create random for hash code.
    rnd = new Random();
    rnd_seed = new Date().getTime();
  }

  public boolean getOK() {
    return ok;
  }

  public URL getURL() {
    return url;
  }

  public File getFile() {
    return file;
  }

  public String getHost() {
    return url.getHost();
  }

  public String getExtension() {
    String s = getURLString();
     if(s.length()==0) return s;
    int idx = s.lastIndexOf('.');
    int idx2 = s.lastIndexOf('/');
    if(idx<0 || idx<idx2 || idx == s.length()-1) return "";
    return s.substring(idx+1,s.length());
  }

  public String getFileName() {
    return file.getName();
  }

  public String getURLString() {
    if(url_str!=null) return url_str;
    String s = url.toString();
    String ref = url.getRef();
    if(ref==null) {
      url_str = s;
      return url_str;
    }
    url_str = s.substring(0,s.length()-ref.length()-1);
    return url_str;
  }

  public String getProtocol() {
    return url.getProtocol();
  }

  public boolean isEndsWithDirectory() {
    return getURLString().endsWith("/");
  }

  public boolean isWithoutExtension() {
    if(!isEndsWithDirectory())
      return getExtension().length()==0;
    else
      return true;
  }

  public int getCount() {
    return counter;
  }

  public boolean getHasNext() {
    return has_next;
  }

  public void setHasNext() {
    has_next = true;
    origin_url = getURLString();
    next_state = 0;
  }

  public void nextURL() {
    try {
      url_str = null;
      if(next_state==default_files.length)
	if(origin_url.endsWith("/")) {
	  url = new URL(url, origin_url);
	  url_str = getURLString();
	  has_next = false;
	  return;
	} else {
	  url = new URL(origin_url+"/");
	  url_str = getURLString();
	  next_state++;
	  return;
	}

      if(next_state==default_files.length+1) {
	url = new URL(origin_url);
	url_str = getURLString();
	file = new File(directory+"/"+url.getHost()+url.getFile());
	has_next = false;
	return;
      }

      if(origin_url.endsWith("/"))
	url = new URL(origin_url+default_files[next_state]);
      else
	url = new URL(origin_url+"/"+default_files[next_state]);
      url_str = getURLString();
      file = new File(directory+"/"+url.getHost()+url.getFile());
      next_state++;

    } catch(MalformedURLException e) {
      System.err.println("RS:MalformedURLException at nextURL "+origin_url);
      ok = false;
    }
  }

  // For Resource class's hash code.
  public Integer getKey() {
    return new Integer(hash(getURLString()));
  }

  private synchronized static int hash(String s) {
    rnd.setSeed(rnd_seed);
    int hash = 0;
    for(int i=0;i<s.length();i++)
      hash = (hash<<1) ^ (rnd.nextInt() * s.charAt(i))& 0xffffffff;
    return hash;
  }
}



