package PetshopSpider;

import java.util.*;
import java.io.*;

public class Spider {
  public static void main(String args[]) {
    // Check file's property.

    Print.message_level = Print.NORMAL;
    File f = new File(args[0]);
    if (!f.canRead()) {
      System.err.println("Error: "+args[0]+" doesn't exists or isn't readable.");
      System.exit(1);
    }

    // Readup properties.
    Properties prop = null;

    try {
      FileInputStream in = new FileInputStream(args[0]); 
      prop = new Properties();
      prop.load(in);
      in.close();
      System.out.println(prop);
    } catch(FileNotFoundException e) {
      System.err.println("SP: Init-file is not found.");
      System.exit(1);
    } catch (IOException e) {
      System.err.println("SP: Init-file is illegal format.");
      System.exit(1);
    } 

    // Run main object.
    System.out.println("SP: Create Spider object");
    new Spider(prop).startSpider();
    
  }

  // Constructor

  Spider(Properties prop) {

    Print.debug("SP: Setup spider setting.");
    // Setup common setting object.
    SpiderSetting sset = new SpiderSetting();
    spider_prop = prop;

    // Timeout.
    sset.timeout = Integer.parseInt(prop.getProperty("timeout","60"));

    // Server retries.
    sset.retries = Integer.parseInt(prop.getProperty("server_retries","5"));
    // Maximum level of URL.
    sset.max_level = Integer.parseInt(prop.getProperty("max_level","40"));
    sset.max_connections = Integer.parseInt(prop.getProperty("max_connections","8"));

    // Number of threads are invoked.
    sset.number_of_retrievers = Integer.parseInt(prop.getProperty("number_of_retrievers","6"));

    // Storage directory name. No default.
    sset.storage_directory = prop.getProperty("storage_directory");

    // Domains. No default.
    sset.domains = getStringArray(prop.getProperty("domains"));

    // Extensions to save files.
    sset.extensions = getStringArray(prop.getProperty("extensions","htm,html"));

    // Extensions not to save files.
    sset.disable_extensions = getStringArray(prop.getProperty("disable_extensions","jpg,gif"));

    // Servers should be ignored.
    sset.ignore_servers = getStringArray(prop.getProperty("ignore_servers",""));

    // If MIME types are included in this list, save it if file name doesn't
    // has enable extensions.
    sset.mimes = getStringArray(prop.getProperty("allow_mime_types",""));

    // Default files.
    sset.default_files = getStringArray(prop.getProperty("default_files","index.html,index.htm"));

    // Robots.txt flag.
    sset.robots_txt = new Boolean(prop.getProperty("robots_txt","true")).booleanValue();

    // Message level.
    String level = prop.getProperty("message_level","normal");
    if(level.compareTo("debug")==0)
      Print.message_level = Print.DEBUG;
    else if(level.compareTo("verbose")==0)
      Print.message_level = Print.VERBOSE;
    else if(level.compareTo("normal")==0)
      Print.message_level = Print.NORMAL;
    else
      Print.message_level = Print.QUIET;

    // Create URLChecker objects.
    url_checker = new URLChecker(sset);

    // Set start URLs.
    String[] s = getStringArray(prop.getProperty("start_urls"));
    sset.urls = new Resource[s.length];

    // Set Resource class's static variable.
    Resource.init(sset);

    for(int i=0;i<sset.urls.length;i++)
      sset.urls[i] = new Resource(s[i], 1);
    job_manager = new JobManager(sset);
    job_manager.setURLChecker(url_checker);
    url_checker.setJobManager(job_manager);

    // Set statics.
    HTMLRetriever.init(sset, job_manager, url_checker);
    WWWHost.init(sset, job_manager);
  }

  private void startSpider() {
    job_manager.start();
    try {
      job_manager.join();
    } catch (InterruptedException e){};
  }
  
  // Parse multiple terms. It converts to lower case.
  private String[] getStringArray(String s) {
    Print.debug("SP:getStringArray "+s);
    StringTokenizer st = new StringTokenizer(s,", ");
    String[] ss = new String[st.countTokens()];
    int i=0;
    while(st.hasMoreTokens()) {
      ss[i++] = st.nextToken().toLowerCase();
    }
    return ss;
  }

  // Variables
  Properties spider_prop;
  JobManager job_manager;
  URLChecker url_checker;
}
