package PetshopSpider;

import java.util.*;
import java.io.*;
import java.net.*;

class URLChecker {

  URLChecker(SpiderSetting sset) {
    host_table = new Hashtable(50);
    url_table = new Hashtable(3000);
    this.timeout = sset.timeout;
    this.domains = sset.domains;
    this.extensions = sset.extensions;
  }

  // Setting JobManager for mutually reference structure.
  public void setJobManager(JobManager jm) {
    job_manager = jm;
  }

  // Host check.
  public synchronized int getHostState(String host) {
    if(host_table.containsKey(host))
      return ((WWWHost)host_table.get(host)).getState();
    else
      return addNewHost(host);
  }

  // Robots.txt check
  public synchronized boolean inRobotsTxt(String host, String url) {
    // Check robots.txt
    return ((WWWHost)host_table.get(host)).inRobotsTxt(url);
  }

  public boolean isAlreadyRetrieved(Integer key) {
    // Check already retrieved or not.
    Print.verbose("UC: url check ");
    return url_table.containsKey(key);
  }

  // Receive URL process finishing report.
  public synchronized void notifyFinish(Integer key) {
    // Add this URL to url_table.
    // Only add hash code of URL string.
    url_table.put(key , new Boolean(true));
  }

  // Add new host. Return true if host is available as a result.
  private synchronized int addNewHost(String host) {
    // Check server's status
    Print.verbose("UC: host check "+host);

    // Check already done.
    if(host_table.containsKey(host)) {
      WWWHost wh = (WWWHost)host_table.get(host);
      return wh.getState();
    }

    // Create new WWWHost.
    Print.message("UC: Adding new host "+host);
    WWWHost wh = new WWWHost(host);
    wh.setPriority(Thread.NORM_PRIORITY);

    // Check robot.txt
    Resource rs = new Resource("http://"+host+"/robots.txt");
    cm = new ConnectionMaker(rs);
    cm.start();
    try {
      cm.join(60000);
    } catch(InterruptedException e) {}
    cm.stop();

    // Check activity of host.
    if(!cm.isSuccess()) {
      Print.message("UC: "+host+" is denied");
      wh.notifyDown();
      host_table.put(host, wh);
      job_manager.addHost(host, wh);
      wh.start();
      return WWWHost.HOST_IS_DOWN;
    }
    
    Print.message("UC: "+host+" is accepted");
   
    // Try to retrieve robots.txt
    String header = cm.getResponse();
    InputStream in = cm.getInputStream();

    if(header==null || in==null) {
      wh.notifyDown();
      host_table.put(host, wh);
      job_manager.addHost(host, wh);
      wh.start();
      Print.message("UC: No HTTP header. "+host+" is down.");
      return WWWHost.HOST_IS_DOWN;
    }

    host_table.put(host, wh);

    // No robots.txt is there.
    Print.verbose("UC: Retrieve robots.txt");
    if(header.indexOf("200")<0) {
      job_manager.addHost(host, wh);
      wh.start(); // Start WWWHost
      return WWWHost.HOST_IS_OK;
    }

    // Analyze robots.txt
    BufferedReader br=null;
    br = new BufferedReader(new InputStreamReader(in));

    Vector disallows = new Vector();
    String user_agent = "*";

    Print.verbose("UC: Start analyzing robots.txt");
    try {
      for(String line = br.readLine();line!=null;line = br.readLine()) {
	Print.verbose("UC: robots.txt line "+line);
	StringTokenizer st = new StringTokenizer(line);

	// Disallow: /... or User-Agent: * type format only.
	if(st.countTokens()<2) continue;

	// Agent or directory string.
	String s = st.nextToken();
	if(s.toLowerCase().compareTo("user-agent:")==0) {
	  s = st.nextToken();
	  // PetshopSpider string is more important than any('*').
	  if(user_agent.compareTo("PetshopSpider")==0) break;
	  user_agent = s;
	  disallows.removeAllElements(); // Clear temporary disallows.
	} else 
	  if(s.toLowerCase().compareTo("disallow:")==0) {
	    // An agent not PetshopSpider description is skipped.
	    if (user_agent.compareTo("*")==0 || user_agent.compareTo("PetshopSpider")==0) {
	      // For /* type description
	      while(st.hasMoreTokens()) {
		s = st.nextToken();
		s = s.replace('*',' ').trim();
		Print.debug("UC: Add a disallow http://"+host+s);
		disallows.addElement("http://"+host+s);
	      }
	    }
	  }
      }
    } catch(IOException e) {}

    try {
      br.close();
    } catch(IOException e) {};

    // Skip no lines case.
    if(disallows.isEmpty()) {
      job_manager.addHost(host, wh);
      wh.start(); // WWWHost started.
      return WWWHost.HOST_IS_OK;
    }

    // Set value.
    wh.setRobotsTxt(disallows);
    job_manager.addHost(host, wh);
    wh.start(); // Start WWWhost.
    return WWWHost.HOST_IS_OK;
  }

  // Variables
  private Hashtable host_table;
  private Hashtable url_table;
  private Hashtable robots_txt;
  private Hashtable no_response;
  private String[] domains, extensions;
  private JobManager job_manager;

  int timeout;
  ConnectionMaker cm;

}





