package PetshopSpider;

import java.util.*;

class WWWHost extends Thread {

  // WWHost constrcutor
  WWWHost(String name) {

    // Initialize private variables.
    continuas_fail = 0; continuas_success = 0;
    has_robots_txt = false;
    connection_limit = max_connections;
    current_connections = 0;
    host_state = HOST_IS_OK;
    queue = new Hashtable();
    id = next_id++; // Assign next ID for this object.
    host_name = name; // Give a host name.
    requested = false; // Inner state of request.
    setDaemon(true);
    receive_resume = false;
  }


  /*
    Static methods. Called by Spider's main methods.
    */

  // Set static variables. (Static method. This information is common over all
  // WWWHost objects.

  public static void init(SpiderSetting sset, JobManager jm) {
    job_manager = jm;
    retries_limit = sset.retries;
    max_connections = sset.number_of_retrievers;
    next_id = 0;
  }

  /*
    run methods for timeout.
    */

  public void run() {
    Print.debug("WH: "+host_name+" start main loop");
    for(;;) {
      job_manager.notifyStable();
      if(!receive_resume) 
	suspend();
      job_manager.notifyInStable();
      receive_resume = false;

      if(host_state==HOST_IS_DOWN) {
	Print.debug("WH: "+host_name+" is down.");
	job_manager.notifyStable();
	this.stop();
      }

      if(host_state==HOST_IS_BUSY) {
	Print.debug("WH: "+host_name+" is busy.");
	job_manager.notifyBusy();
	job_manager.notifyStable(); // Not to prevent other WWWHosts.
	try {
	  Thread.sleep(60000);
	} catch(InterruptedException e) {}
	host_state = HOST_IS_OK;
	job_manager.notifyInStable(); // Not to prevent other WWWHosts.
	job_manager.recoverFromBusy();
	/*
	  Here is no continue because that it confuses JobManager thinks
	  all jobs are stopped although queue is not empty.
	  */
      }

      if(queue.isEmpty()) {
	Print.debug("WH: "+host_name+"'s queue is empty.");
	if(requested) {
	  job_manager.cancelHR(id);
	  requested = false;
	}
	continue;
      }

      if(current_connections>=connection_limit) {
	if(requested) {
	  job_manager.cancelHR(id);
	  requested = false;
	}
	Print.verbose("WH: "+host_name+"'s connections exceeds limit");
	continue;
      }

      Print.debug("WH: Request HR ID:"+id);
      if(!requested) {
	requested = true;
	Print.debug("WH: Call RequestHR ID:"+id);
	job_manager.requestHR(id);
      }
    }
  }

  // For debug.
  public String getHostName() {
    return host_name;
  }

  /*
    Public methdos used by JobManager.
    */

  // Adding a job into queue.
  public synchronized void addJob(Resource rs) {
    Print.debug("WH: addJob is called for "+rs.getURLString()+", size is "+queue.size());
    if(host_state==HOST_IS_DOWN) return;
    Integer key = rs.getKey();
    if(!queue.containsKey(key))
       queue.put(key,rs);
    receive_resume = true;
    this.resume();
  }

  // Send a job to HTMLretriever. True for accept, false for reject.
  public synchronized void giveHR(HTMLRetriever hr) {
    if(host_state==HOST_IS_DOWN) return;
    Print.debug("WH: giveHR is called, "+current_connections+" connection already exist for "+host_name);

    // Send a job.
    Enumeration e = queue.keys();
    Integer key = (Integer)e.nextElement();
    Resource rs = (Resource)queue.get(key);
    queue.remove(key);
    current_connections++;
    requested = false;
    hr.setJob(this, rs); // HTMLRetriever callbacks to change # of connections.

    // For next job.
    receive_resume = true;
    this.resume();
  }

  /*
    Public methods used by URLChecker.
    */

  // Set robots.txt
  public synchronized void setRobotsTxt(Vector txt) {
    robots_txt = txt;
    has_robots_txt = true;
  }

  // Check robots.txt
  public synchronized boolean inRobotsTxt(String url) {
    // If the host has no robots.txt, returns.
    if(!has_robots_txt) return false;

    // Check robots.txt corresponding to URL.
    Enumeration e = robots_txt.elements();
    while(e.hasMoreElements())
      if(url.startsWith((String)e.nextElement())) return true;
    return false;
  }

  // Get current host state.
  public int getState() {
    return host_state;
  }

  public synchronized void notifyDown() {
    host_state = HOST_IS_DOWN;
  }

  /*
    public method caled by HTMLretriever
    */
    
  public synchronized void notifyFinish(int result) {
    current_connections--;

    // Connection control.
    switch(result) {
    case RETRIEVE_FAIL: {

      if(host_state!=HOST_IS_OK)
	return;

      continuas_success = 0;
      continuas_fail++;

      if(host_state != HOST_IS_OK) return;

      // Cancel request.
      if(requested) {
	job_manager.cancelHR(id);
	requested = false;
      }

      if(continuas_fail>retries_limit) {
	Print.message("WH: "+host_name+" is down");
	host_state = HOST_IS_DOWN;
	return;
      }

      host_state = HOST_IS_BUSY;
      
      // Decrease connection's limit.
      Print.message("WH: No response from "+host_name+", decrease connection limit");
      connection_limit = (int)(((float)connection_limit)/2.0);
      if(connection_limit<1) connection_limit = 1;
      break;
    }
    
    case RETRIEVE_SUCCESS: {
      if(++continuas_success>INCREASE_LIMIT) {
	connection_limit++;
	continuas_success = 0;
      }
      continuas_fail =0;
      break;
    }

    }

    receive_resume = true;
    this.resume();
  }

  // Variables
  private int host_state;
  // For controlling connection increase, decrease.
  private int continuas_fail, continuas_success;
  private Vector robots_txt;
  private boolean has_robots_txt;
  private int connection_limit, current_connections;
  private Hashtable queue;
  private String host_name;
  
  private boolean requested, receive_resume;
  private int id; // The unique integer for objects.

  private static int retries_limit;
  private static int max_connections;
  private static JobManager job_manager;
  private static int next_id;

  public final static int HOST_IS_OK = 0;
  public final static int HOST_IS_BUSY = 1;
  public final static int HOST_IS_DOWN = 2;

  // Constants
  public final static int RETRIEVE_SKIP = 0;
  public final static int RETRIEVE_FAIL = 1;
  public final static int RETRIEVE_SUCCESS = 2;

  // The condition for increase connection limit.
  public final static int INCREASE_LIMIT = 10;
}
